/*
 * Copyright (C) 2002 2003 2005, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "SDL.h"
#include "sound.h" 
#include "ringbuf.h"

static struct {
     Ringbuf *output_buffer;
     gfloat delay_time;
     off_t played_bytes;
     guint32 delay_bytes;
} sdl_data;

static void sdl_init(void)
{
     if (SDL_Init(SDL_INIT_AUDIO) == -1) {
	  printf("Could not initialize SDL: %s.\n", SDL_GetError());
	  exit(1);
     }
     sdl_data.output_buffer = ringbuf_new ( 
	  inifile_get_guint32(INI_SETTING_SOUNDBUFSIZE,
                              INI_SETTING_SOUNDBUFSIZE_DEFAULT) );
}

static void sdl_quit(void)
{
     SDL_Quit();
}

static void sdl_output_callback(void *userdata, Uint8 *stream, int len)
{
     guint32 ui;
     /* printf("output_callback: before: %d,%d\n",
	ringbuf_available(sdl_data.output_buffer),
	ringbuf_freespace(sdl_data.output_buffer)); */
     ui = ringbuf_dequeue ( sdl_data.output_buffer, stream, len );
     /*printf("output_callback: after: %d,%d (%d)\n",
       ringbuf_available(sdl_data.output_buffer),
       ringbuf_freespace(sdl_data.output_buffer),ui); */
     sdl_data.played_bytes += ui;
}

static gboolean sdl_output_select_format(Dataformat *format)
{
     SDL_AudioSpec desired;
     if (format->type == DATAFORMAT_FLOAT || format->samplesize > 2 || 
	 format->channels > 2) return TRUE;
     desired.freq = format->samplerate;
     if (format->sign) {
	  if (format->samplesize == 1)
	       desired.format = AUDIO_S8;
	  else if (format->bigendian)
	       desired.format = AUDIO_S16MSB;
	  else
	       desired.format = AUDIO_S16LSB;
     } else {
	  if (format->samplesize == 1)
	       desired.format = AUDIO_U8;
	  else if (format->bigendian)
	       desired.format = AUDIO_U16MSB;
	  else
	       desired.format = AUDIO_U16LSB;
     }
     desired.channels = format->channels;
     desired.samples = 512;
     desired.callback = sdl_output_callback;
     if (SDL_OpenAudio(&desired,NULL) < 0) {
	  fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
	  return TRUE;
     }
     sdl_data.delay_time=
	  (gfloat)((desired.samples)/format->samplebytes)/
	  (gfloat)format->samplerate;
     sdl_data.delay_bytes = 512; /* ?? */
     sdl_data.played_bytes = 0;
     return FALSE;
}

static void sdl_output_stop(void)
{
     // puts("output_stop");
     while (ringbuf_available(sdl_data.output_buffer) > 0) do_yield(TRUE);
     if (SDL_GetAudioStatus() != SDL_AUDIO_STOPPED) 
	  SDL_CloseAudio();
     ringbuf_drain(sdl_data.output_buffer);
}

static void sdl_output_clear_buffers(void)
{
     SDL_PauseAudio(1);
     ringbuf_drain(sdl_data.output_buffer);
}

static guint sdl_output_play(gchar *buffer, guint bufsize)
{
     guint i;
     if (!bufsize) return ringbuf_available(sdl_data.output_buffer);
     SDL_LockAudio();
     /* printf("output_play: before: %d,%d\n",ringbuf_available(
	sdl_data.output_buffer),ringbuf_freespace(sdl_data.output_buffer)); */
     i = (guint)ringbuf_enqueue( sdl_data.output_buffer, buffer, bufsize );
     /* printf("output_play: after: %d,%d\n",
	ringbuf_available(sdl_data.output_buffer),
	ringbuf_freespace(sdl_data.output_buffer)); */
     SDL_UnlockAudio();
     if (SDL_GetAudioStatus() == SDL_AUDIO_PAUSED)
	  SDL_PauseAudio(0);
     return i;
}

static gboolean sdl_output_want_data(void)
{
     return !ringbuf_isfull(sdl_data.output_buffer);
}

static gfloat sdl_output_play_min_delay_time(void)
{
     return sdl_data.delay_time;
}

static off_t sdl_output_played_bytes(void)
{
     return (sdl_data.played_bytes > sdl_data.delay_bytes) ?
	  (sdl_data.played_bytes - sdl_data.delay_bytes) : 0;
}
